
cimport cython
import numpy as np
cimport numpy as np

from libc.math cimport log, sqrt


def cython_fit_BlackScholesMerton_model(np.ndarray[double, ndim=1] ts, np.ndarray[double, ndim=1] prices):
    cdef int nbpts = len(prices)
    cdef int i
    cdef double dlogS, dt

    cdef double sumpr = 0.
    cdef double sumnoise = 0.
    cdef double sumsqnoise = 0.

    for i in range(nbpts-1):
        dlogS = log(prices[i+1]/prices[i])
        dt = ts[i+1] - ts[i]

        sumpr += dlogS / dt
        sumnoise += dlogS / sqrt(dt)
        sumsqnoise += dlogS * dlogS / dt

    cdef double r = sumpr / (nbpts - 1)
    cdef double sigma = sqrt(sumsqnoise / (nbpts-1) - sumnoise*sumnoise/(nbpts-1)/(nbpts-1))

    return r, sigma


def cython_fit_multivariate_BlackScholesMerton_model(np.ndarray[double, ndim=1] ts, np.ndarray[double, ndim=2] multiprices):
    cdef np.ndarray[double, ndim=2] dlogS = np.log(multiprices[:, 1:] / multiprices[:, :-1])
    cdef np.ndarray[double, ndim=1] dt = ts[1:] - ts[:-1]

    cdef np.ndarray[double, ndim=1] r = np.zeros(multiprices.shape[0])
    for i in range(multiprices.shape[0]):
        r[i] = np.mean((dlogS / dt)[i, :])
    cdef np.ndarray[double, ndim=2] cov = np.cov(dlogS / np.sqrt(dt))

    return r, cov




import unittest

import numpy as np

from finsim.portfolio.optimize import metrics


class TestMetrics(unittest.TestCase):
    def test_sharpe_ratio(self):
        weights = np.array([0.3, 0.5, 0.2])
        r = np.array([0.1, 0.05, 0.15])
        cov = np.array([[1., 0.4, -0.3], [0.4, 1., 0.05], [-0.3, 0.05, 1.]])
        rf = 0.001
        expected_sharpe_ratio = 0.12200850769256662

        self.assertAlmostEqual(
            metrics.sharpe_ratio(weights, r, cov, rf, lowlevellang='C'),
            expected_sharpe_ratio,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.sharpe_ratio(weights, r, cov, rf, lowlevellang='P'),
            expected_sharpe_ratio,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.sharpe_ratio(weights, r, cov, rf, lowlevellang='F'),
            expected_sharpe_ratio,
            delta=1e-6
        )

    def test_costfunction(self):
        weights = np.array([0.05, 0.3, 0.5, 0.2])
        r = np.array([0.1, 0.05, 0.15])
        cov = np.array([[1., 0.4, -0.3], [0.4, 1., 0.05], [-0.3, 0.05, 1.]])
        rf = 0.001

        lamb01cost = 0.077475
        self.assertAlmostEqual(
            metrics.mpt_costfunction(weights, r, cov, rf, 0.1, V0=10, lowlevellang='C'),
            lamb01cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_costfunction(weights, r, cov, rf, 0.1, V0=10, lowlevellang='P'),
            lamb01cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_costfunction(weights, r, cov, rf, 0.1, V0=10, lowlevellang='F'),
            lamb01cost,
            delta=1e-6
        )

        lamb02cost = 0.059750000000000004
        self.assertAlmostEqual(
            metrics.mpt_costfunction(weights, r, cov, rf, 0.2, V0=10, lowlevellang='C'),
            lamb02cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_costfunction(weights, r, cov, rf, 0.2, V0=10, lowlevellang='P'),
            lamb02cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_costfunction(weights, r, cov, rf, 0.2, V0=10, lowlevellang='F'),
            lamb02cost,
            delta=1e-6
        )

    def test_entropycostfunction(self):
        weights = np.array([0.05, 0.3, 0.5, 0.2])
        r = np.array([0.1, 0.05, 0.15])
        cov = np.array([[1., 0.4, -0.3], [0.4, 1., 0.05], [-0.3, 0.05, 1.]])
        rf = 0.001

        case1cost = 0.08170682914813143
        self.assertAlmostEqual(
            metrics.mpt_entropy_costfunction(weights, r, cov, rf, 0.1, 0.01, V=10, lowlevellang='C'),
            case1cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_entropy_costfunction(weights, r, cov, rf, 0.1, 0.01, V=10, lowlevellang='P'),
            case1cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_entropy_costfunction(weights, r, cov, rf, 0.1, 0.01, V=10, lowlevellang='F'),
            case1cost,
            delta=1e-6
        )

        case2cost = 0.11979329148131437
        self.assertAlmostEqual(
            metrics.mpt_entropy_costfunction(weights, r, cov, rf, 0.1, 0.1, V=10, lowlevellang='C'),
            case2cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_entropy_costfunction(weights, r, cov, rf, 0.1, 0.1, V=10, lowlevellang='P'),
            case2cost,
            delta=1e-6
        )
        self.assertAlmostEqual(
            metrics.mpt_entropy_costfunction(weights, r, cov, rf, 0.1, 0.1, V=10, lowlevellang='F'),
            case2cost,
            delta=1e-6
        )


if __name__ == '__main__':
    unittest.main()




from .optimize.numerics import get_BlackScholesMerton_stocks_estimation
from .portfolio import OptimizedPortfolio
from .optimize.policy import OptimizedWeightingPolicyUsingMPTSharpeRatio, \
    OptimizedWeightingPolicyUsingMPTCostFunction, OptimizedWeightingPolicyUsingMPTEntropyCostFunction


def get_optimized_portfolio_on_sharpe_ratio(
        rf,
        symbols,
        totalworth,
        presetdate,
        estimating_startdate,
        estimating_enddate,
        minweight=0.,
        lazy=False,
        cacheddir=None,
        include_dividends=False
):
    r, cov = get_BlackScholesMerton_stocks_estimation(
        symbols,
        estimating_startdate,
        estimating_enddate,
        cacheddir=cacheddir,
        include_dividends=include_dividends
    )
    optimized_weighting_policy = OptimizedWeightingPolicyUsingMPTSharpeRatio(rf, r, cov, symbols, minweight=minweight)
    optimized_portfolio = OptimizedPortfolio(optimized_weighting_policy, totalworth, presetdate, cacheddir=cacheddir)
    return optimized_portfolio


def get_optimized_portfolio_on_mpt_costfunction(
        rf,
        symbols,
        totalworth,
        presetdate,
        estimating_startdate,
        estimating_enddate,
        lamb,
        V0=10.,
        cacheddir=None,
        include_dividends=False
):
    r, cov = get_BlackScholesMerton_stocks_estimation(
        symbols,
        estimating_startdate,
        estimating_enddate,
        cacheddir=cacheddir,
        include_dividends=include_dividends
    )
    optimized_weighting_policy = OptimizedWeightingPolicyUsingMPTCostFunction(rf, r, cov, symbols, lamb, V0=V0)
    optimized_portfolio = OptimizedPortfolio(optimized_weighting_policy, totalworth, presetdate, cacheddir=cacheddir)
    return optimized_portfolio


def get_optimized_portfolio_on_mpt_entropy_costfunction(
        rf,
        symbols,
        totalworth,
        presetdate,
        estimating_startdate,
        estimating_enddate,
        lamb0,
        lamb1,
        V=10.,
        cacheddir=None,
        include_dividends=False
):
    r, cov = get_BlackScholesMerton_stocks_estimation(
        symbols,
        estimating_startdate,
        estimating_enddate,
        cacheddir=cacheddir,
        include_dividends=include_dividends
    )
    optimized_weighting_policy = OptimizedWeightingPolicyUsingMPTEntropyCostFunction(rf, r, cov, symbols, lamb0, lamb1, V=V)
    optimized_portfolio = OptimizedPortfolio(optimized_weighting_policy, totalworth, presetdate, cacheddir=cacheddir)
    return optimized_portfolio




import numpy as np

from .constants import dividing_factors_dict
from .native.pyfit import python_fit_BlackScholesMerton_model, python_fit_multivariate_BlackScholesMerton_model
from .native.cythonfit import cython_fit_BlackScholesMerton_model, cython_fit_multivariate_BlackScholesMerton_model
from .native.fortranfit import fortranfit


# Note: always round-off to seconds first, but flexible about the unit to be used.

def fit_BlackScholesMerton_model(timestamps, prices, unit='year', lowlevellang='F'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    if lowlevellang == 'C':
        return cython_fit_BlackScholesMerton_model(ts, prices)
    elif lowlevellang == 'P':
        return python_fit_BlackScholesMerton_model(ts, prices)
    elif lowlevellang == 'F':
        results = fortranfit.f90_fit_blackscholesmerton_model(ts, prices)
        return results[0], results[1]
    else:
        raise ValueError(
            'Unknown low-level language: {}. (Should be "P" (Python), "C" (Cython), or "F" (Fortran).)'.format(
                lowlevellang))


def fit_multivariate_BlackScholesMerton_model(timestamps, multiprices, unit='year', lowlevellang='C'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    if lowlevellang == 'C':
        return cython_fit_multivariate_BlackScholesMerton_model(ts, multiprices)
    elif lowlevellang == 'N':
        return python_fit_multivariate_BlackScholesMerton_model(ts, multiprices)
    else:
        raise ValueError(
            'Unknown low-level language: {}. (Should be "C" (Cython), or "F" (Fortran).)'.format(
                lowlevellang))




import numpy as np
cimport numpy as np

from libc.math cimport log, sqrt


def cython_estimate_downside_risk(np.ndarray[double, ndim=1] ts, np.ndarray[double, ndim=1] prices, double target_return):
    cdef int nbpts = len(prices)
    cdef double logS, dt, rms, sum_down_sqrms
    cdef int i

    sum_down_sqrms = 0.
    for i in range(nbpts-1):
        dlogS = log(prices[i+1] / prices[i])
        dt = ts[i+1] - ts[i]
        rms = dlogS / sqrt(dt)
        if rms < target_return:
            sum_down_sqrms += (target_return-rms) * (target_return-rms)

    cdef downside_risk = sqrt(sum_down_sqrms / (nbpts-1))

    return downside_risk


def cython_estimate_upside_risk(np.ndarray[double, ndim=1] ts, np.ndarray[double, ndim=1] prices, double target_return):
    cdef int nbpts = len(prices)
    cdef double logS, dt, rms, sum_up_sqrms
    cdef int i

    sum_up_sqrms = 0.
    for i in range(nbpts-1):
        dlogS = log(prices[i+1] / prices[i])
        dt = ts[i+1] - ts[i]
        rms = dlogS / sqrt(dt)
        if rms > target_return:
            sum_up_sqrms += (rms-target_return) * (rms-target_return)

    cdef upside_risk = sqrt(sum_up_sqrms / (nbpts-1))

    return upside_risk




from .portfolio import Portfolio, OptimizedPortfolio
from .dynamic import DynamicPortfolio, DynamicPortfolioWithDividends, InsufficientSharesException
from .create import get_optimized_portfolio_on_sharpe_ratio, get_optimized_portfolio_on_mpt_costfunction, \
    get_optimized_portfolio_on_mpt_entropy_costfunction
from .optimize.numerics import get_BlackScholesMerton_stocks_estimation




# Reference: https://stackoverflow.com/questions/7932028/setup-py-for-packages-that-depend-on-both-cython-and-f2py

from setuptools import setup

try:
    from Cython.Build import cythonize
    ext_modules = cythonize(['finsim/portfolio/optimize/native/cythonmetrics.pyx',
                             'finsim/estimate/native/cythonfit.pyx',
                             'finsim/estimate/native/cythonrisk.pyx'])
except ImportError:
    from setuptools import Extension
    ext_modules = [
        Extension('finsim.portfolio.optimize.native.cynthonmetrics',
                  sources=['finsim/portfolio/optimize/native/cythonmetrics.c']),
        Extension('finsim.estimate.native.cythonfit',
                  sources=['finsim/estimate/native/cythonfit.c']),
        Extension('finsim.estimate.native.cythonrisk',
                  sources=['finsim/estimate/native/cythonrisk.c'])
    ]

import numpy as np
from numpy.distutils.core import setup
from numpy.distutils.core import Extension as fortranExtension

fortran_ext_modules = [
    fortranExtension(
        'finsim.portfolio.optimize.native.fortranmetrics',
        sources=['finsim/portfolio/optimize/native/fortranmetrics.f90',
                 'finsim/portfolio/optimize/native/fortranmetrics.pyf']
    ),
    fortranExtension(
        'finsim.estimate.native.fortranfit',
        sources=['finsim/estimate/native/fortranfit.f90',
                 'finsim/estimate/native/fortranfit.pyf']
    ),
    fortranExtension(
        'finsim.estimate.native.fortranrisk',
        sources=['finsim/estimate/native/fortranrisk.f90',
                 'finsim/estimate/native/fortranrisk.pyf']
    ),
    fortranExtension(
        'finsim.simulation.native.f90brownian',
        sources=['finsim/simulation/native/brownian.f90',
                 'finsim/simulation/native/brownian.pyf']
    )
]


def readme():
    with open('README.md') as f:
        return f.read()


def install_requirements():
    return [package_string.strip() for package_string in open('requirements.txt', 'r')]


def package_description():
    text = open('README.md', 'r').read()
    startpos = text.find('## Introduction')
    return text[startpos:]


setup(
    name='finsim',
    version="0.7.6",
    description="Financial simulation and inference",
    long_description=package_description(),
    long_description_content_type='text/markdown',
    classifiers=[
      "Topic :: Scientific/Engineering :: Mathematics",
      "Topic :: Software Development :: Libraries :: Python Modules",
      "Topic :: Software Development :: Version Control :: Git",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Cython",
      "Programming Language :: C",
      "Programming Language :: Fortran",
      "Intended Audience :: Science/Research",
      "Intended Audience :: Developers",
      "Intended Audience :: Financial and Insurance Industry"
    ],
    keywords="simulation, finance, quantitative finance, inference, portfolio analysis",
    url="https://github.com/stephenhky/FinanceSimulation",
    author="Kwan-Yuet Ho",
    author_email="stephenhky@yahoo.com.hk",
    license='LGPL',
    packages=[
        'finsim',
        'finsim.data',
        'finsim.estimate',
        'finsim.estimate.native',
        'finsim.simulation',
        'finsim.portfolio',
        'finsim.portfolio.optimize',
        'finsim.portfolio.optimize.native'
    ],
    include_dirs=[np.get_include()],
    setup_requires=['Cython', 'numpy', ],
    install_requires=install_requirements(),
    tests_require=[
      'unittest2'
    ],
    # scripts=[],
    include_package_data=True,
    ext_modules=fortran_ext_modules+ext_modules,
    test_suite="test",
    zip_safe=False
)




import unittest

from finsim.data.preader import get_symbol_closing_price


class TestDynamicPortfolio(unittest.TestCase):
    def test_get_voo_backtrack(self):
        self.assertAlmostEqual(
            get_symbol_closing_price('VOO', '2016-01-01', backtrack=True),
            186.92999267578125
        )

    def test_get_voo_nobacktrack(self):
        self.assertRaises(
            IndexError,
            get_symbol_closing_price,
            'VOO',
            '2016-01-01',
            backtrack=False
        )
        self.assertAlmostEqual(
            get_symbol_closing_price('VOO', '2015-12-31', backtrack=True),
            186.92999267578125
        )


if __name__ == '__main__':
    unittest.main()
